@@ -10,14 +10,26 @@ import io
 import os
 import re
 import struct
+import tempfile
+from typing import IO, List, Optional, Tuple, Type
+
+from intelhex import HexRecordError, IntelHex
 
 from .loader import ESPLoader
 from .targets import (
     ESP32C2ROM,
     ESP32C3ROM,
+    ESP32C5ROM,
+    ESP32C5BETA3ROM,
     ESP32C6BETAROM,
+    ESP32C6ROM,
+    ESP32C61ROM,
     ESP32H2BETA1ROM,
     ESP32H2BETA2ROM,
+    ESP32H2ROM,
+    ESP32H21ROM,
+    ESP32H4ROM,
+    ESP32P4ROM,
     ESP32ROM,
     ESP32S2ROM,
     ESP32S3BETA2ROM,
@@ -33,21 +45,63 @@ def align_file_position(f, size):
     f.seek(align, 1)
 
 
-class ESPBOOTLOADER(object):
-    """
-    These are constants related to software ESP8266 bootloader,
-    working with 'v2' image files
-    """
+def _find_subsequences(addresses: List[int]) -> List[Tuple[int, int]]:
+    """Find continuous subsequences in a list of addresses"""
+    if not addresses:
+        return []
 
-    # First byte of the "v2" application image
-    IMAGE_V2_MAGIC = 0xEA
+    sorted_seq = sorted(addresses)
 
-    # First 'segment' value in a "v2" application image,
-    # appears to be a constant version value?
-    IMAGE_V2_SEGMENT = 4
+    subsequences = []
+    start = sorted_seq[0]
+
+    for prev, num in zip(sorted_seq, sorted_seq[1:]):
+        if num != prev + 1:
+            # Found a gap, save the current subsequence
+            subsequences.append((start, prev))
+            start = num
+
+    # Add the last subsequence
+    subsequences.append((start, sorted_seq[-1]))
+
+    return subsequences
 
 
-def LoadFirmwareImage(chip, filename):
+def _split_intel_hex_file(ih: IntelHex) -> List[Tuple[int, IO[bytes]]]:
+    """Split an IntelHex file into multiple temporary binary files based on the gaps
+    in the addresses"""
+    subsequences = _find_subsequences(ih.addresses())
+    bins: List[Tuple[int, IO[bytes]]] = []
+    for start, end in subsequences:
+        bin = tempfile.NamedTemporaryFile(suffix=".bin", delete=False)
+        ih.tobinfile(bin, start=start, end=end)
+        bin.seek(0)  # make sure the file is at the beginning
+        bins.append((start, bin))
+    return bins
+
+
+def intel_hex_to_bin(
+    file: IO[bytes], start_addr: Optional[int] = None
+) -> List[Tuple[Optional[int], IO[bytes]]]:
+    """Convert IntelHex file to list of temp binary files
+    If not hex file return input file otherwise"""
+    INTEL_HEX_MAGIC = b":"
+    magic = file.read(1)
+    file.seek(0)
+    try:
+        if magic == INTEL_HEX_MAGIC:
+            ih = IntelHex()
+            ih.loadhex(file.name)
+            file.close()
+            return _split_intel_hex_file(ih)  # type: ignore
+        else:
+            return [(start_addr, file)]
+    except (HexRecordError, UnicodeDecodeError):
+        # file started with HEX magic but the rest was not according to the standard
+        return [(start_addr, file)]
+
+
+def LoadFirmwareImage(chip, image_file):
     """
     Load a firmware image. Can be for any supported SoC.
 
@@ -57,8 +111,9 @@ def LoadFirmwareImage(chip, filename):
     Returns a BaseFirmwareImage subclass, either ESP8266ROMFirmwareImage (v1)
     or ESP8266V2FirmwareImage (v2).
     """
-    chip = re.sub(r"[-()]", "", chip.lower())
-    with open(filename, "rb") as f:
+
+    def select_image_class(f, chip):
+        chip = re.sub(r"[-()]", "", chip.lower())
         if chip != "esp8266":
             return {
                 "esp32": ESP32FirmwareImage,
@@ -70,17 +125,30 @@ def LoadFirmwareImage(chip, filename):
                 "esp32h2beta1": ESP32H2BETA1FirmwareImage,
                 "esp32h2beta2": ESP32H2BETA2FirmwareImage,
                 "esp32c2": ESP32C2FirmwareImage,
+                "esp32c6": ESP32C6FirmwareImage,
+                "esp32c61": ESP32C61FirmwareImage,
+                "esp32c5": ESP32C5FirmwareImage,
+                "esp32c5beta3": ESP32C5BETA3FirmwareImage,
+                "esp32h2": ESP32H2FirmwareImage,
+                "esp32h21": ESP32H21FirmwareImage,
+                "esp32p4": ESP32P4FirmwareImage,
+                "esp32h4": ESP32H4FirmwareImage,
             }[chip](f)
         else:  # Otherwise, ESP8266 so look at magic to determine the image type
             magic = ord(f.read(1))
             f.seek(0)
             if magic == ESPLoader.ESP_IMAGE_MAGIC:
                 return ESP8266ROMFirmwareImage(f)
-            elif magic == ESPBOOTLOADER.IMAGE_V2_MAGIC:
+            elif magic == ESP8266V2FirmwareImage.IMAGE_V2_MAGIC:
                 return ESP8266V2FirmwareImage(f)
             else:
                 raise FatalError("Invalid image magic number: %d" % magic)
 
+    if isinstance(image_file, str):
+        with open(image_file, "rb") as f:
+            return select_image_class(f, chip)
+    return select_image_class(image_file, chip)
+
 
 class ImageSegment(object):
     """Wrapper class for a segment in an ESP image
@@ -149,6 +217,8 @@ class ELFSection(ImageSegment):
 class BaseFirmwareImage(object):
     SEG_HEADER_LEN = 8
     SHA256_DIGEST_LEN = 32
+    IROM_ALIGN = 0
+    MMU_PAGE_SIZE_CONF: Tuple[int, ...] = ()
 
     """ Base class with common firmware image functions """
 
@@ -157,6 +227,7 @@ class BaseFirmwareImage(object):
         self.entrypoint = 0
         self.elf_sha256 = None
         self.elf_sha256_offset = 0
+        self.pad_to_size = 0
 
     def load_common_header(self, load_file, expected_magic):
         (
@@ -240,17 +311,43 @@ class BaseFirmwareImage(object):
             )
         return segment_data
 
-    def save_segment(self, f, segment, checksum=None):
+    def save_segment(self, f, segment, checksum=None, segment_name=None):
         """
         Save the next segment to the image file,
         return next checksum value if provided
         """
         segment_data = self.maybe_patch_segment_data(f, segment.data)
-        f.write(struct.pack("<II", segment.addr, len(segment_data)))
+        segment_len = len(segment_data)
+        segment_name = segment_name if segment_name is not None else ""
+        if segment_len & 3:
+            raise FatalError(
+                f"Invalid {segment_name} segment length {segment_len:#x}. It has to be multiple of 4."
+            )
+        SIXTEEN_MB = 0x1000000
+        if segment_len >= SIXTEEN_MB:
+            raise FatalError(
+                f"Invalid {segment_name} segment length {segment_len:#x}. The 16 MB limit has been exceeded."
+            )
+        f.write(struct.pack("<II", segment.addr, segment_len))
         f.write(segment_data)
         if checksum is not None:
             return ESPLoader.checksum(segment_data, checksum)
 
+    def save_flash_segment(self, f, segment, checksum=None):
+        """
+        Save the next segment to the image file, return next checksum value if provided
+        """
+        if self.ROM_LOADER.CHIP_NAME == "ESP32":
+            # Work around a bug in ESP-IDF 2nd stage bootloader, that it didn't map the
+            # last MMU page, if an IROM/DROM segment was < 0x24 bytes
+            # over the page boundary.
+            segment_end_pos = f.tell() + len(segment.data) + self.SEG_HEADER_LEN
+            segment_len_remainder = segment_end_pos % self.IROM_ALIGN
+            if segment_len_remainder < 0x24:
+                segment.data += b"\x00" * (0x24 - segment_len_remainder)
+        segment_name = getattr(segment, "name", None)
+        return self.save_segment(f, segment, checksum, segment_name)
+
     def read_checksum(self, f):
         """Return ESPLoader checksum from end of just-read image"""
         # Skip the padding. The checksum is stored in the last byte so that the
@@ -308,6 +405,11 @@ class BaseFirmwareImage(object):
         irom_segment = self.get_irom_segment()
         return [s for s in self.segments if s != irom_segment]
 
+    def sort_segments(self):
+        if not self.segments:
+            return  # nothing to sort
+        self.segments = sorted(self.segments, key=lambda s: s.addr)
+
     def merge_adjacent_segments(self):
         if not self.segments:
             return  # nothing to merge
@@ -350,14 +452,23 @@ class BaseFirmwareImage(object):
         self.segments = segments
 
     def set_mmu_page_size(self, size):
-        """
-        If supported, this should be overridden by the chip-specific class.
-        Gets called in elf2image.
-        """
-        print(
-            "WARNING: Changing MMU page size is not supported on {}! "
-            "Defaulting to 64KB.".format(self.ROM_LOADER.CHIP_NAME)
-        )
+        """Set the MMU page size for the image if supported by the chip."""
+        if not self.MMU_PAGE_SIZE_CONF and size != self.IROM_ALIGN:
+            # For chips where MMU page size cannot be set or is fixed, just log a warning and use default if there is one.
+            print(
+                f"Warning: Changing MMU page size is not supported on {self.ROM_LOADER.CHIP_NAME}! "
+                f"Defaulting to {self.IROM_ALIGN // 1024}KB."
+                if self.IROM_ALIGN != 0
+                else ""
+            )
+        elif self.MMU_PAGE_SIZE_CONF and size not in self.MMU_PAGE_SIZE_CONF:
+            # For chips with configurable MMU page sizes, error is raised when the size is not valid.
+            valid_sizes = ", ".join(f"{x // 1024}KB" for x in self.MMU_PAGE_SIZE_CONF)
+            raise FatalError(
+                f"{size} bytes is not a valid {self.ROM_LOADER.CHIP_NAME} page size, select from {valid_sizes}."
+            )
+        else:
+            self.IROM_ALIGN = size
 
 
 class ESP8266ROMFirmwareImage(BaseFirmwareImage):
@@ -415,13 +526,19 @@ class ESP8266V2FirmwareImage(BaseFirmwareImage):
     """
 
     ROM_LOADER = ESP8266ROM
+    # First byte of the "v2" application image
+    IMAGE_V2_MAGIC = 0xEA
+
+    # First 'segment' value in a "v2" application image,
+    # appears to be a constant version value?
+    IMAGE_V2_SEGMENT = 4
 
     def __init__(self, load_file=None):
         super(ESP8266V2FirmwareImage, self).__init__()
         self.version = 2
         if load_file is not None:
-            segments = self.load_common_header(load_file, ESPBOOTLOADER.IMAGE_V2_MAGIC)
-            if segments != ESPBOOTLOADER.IMAGE_V2_SEGMENT:
+            segments = self.load_common_header(load_file, self.IMAGE_V2_MAGIC)
+            if segments != self.IMAGE_V2_SEGMENT:
                 # segment count is not really segment count here,
                 # but we expect to see '4'
                 print(
@@ -489,8 +606,8 @@ class ESP8266V2FirmwareImage(BaseFirmwareImage):
             f.write(
                 struct.pack(
                     b"<BBBBI",
-                    ESPBOOTLOADER.IMAGE_V2_MAGIC,
-                    ESPBOOTLOADER.IMAGE_V2_SEGMENT,
+                    self.IMAGE_V2_MAGIC,
+                    self.IMAGE_V2_SEGMENT,
                     self.flash_mode,
                     self.flash_size_freq,
                     self.entrypoint,
@@ -537,10 +654,10 @@ class ESP32FirmwareImage(BaseFirmwareImage):
     """ESP32 firmware image is very similar to V1 ESP8266 image,
     except with an additional 16 byte reserved header at top of image,
     and because of new flash mapping capabilities the flash-mapped regions
-    can be placed in the normal image (just @ 64kB padded offsets).
+    can be placed in the normal image (just @ MMU page size padded offsets).
     """
 
-    ROM_LOADER = ESP32ROM
+    ROM_LOADER: Type[ESPLoader] = ESP32ROM
 
     # ROM bootloader will read the wp_pin field if SPI flash
     # pins are remapped via flash. IDF actually enables QIO only
@@ -548,11 +665,11 @@ class ESP32FirmwareImage(BaseFirmwareImage):
     # to be set to this value so ROM bootloader will skip it.
     WP_PIN_DISABLED = 0xEE
 
-    EXTENDED_HEADER_STRUCT_FMT = "<BBBBHB" + ("B" * 8) + "B"
+    EXTENDED_HEADER_STRUCT_FMT = "<BBBBHBHH" + ("B" * 4) + "B"
 
     IROM_ALIGN = 65536
 
-    def __init__(self, load_file=None, append_digest=True):
+    def __init__(self, load_file=None, append_digest=True, ram_only_header=False):
         super(ESP32FirmwareImage, self).__init__()
         self.secure_pad = None
         self.flash_mode = 0
@@ -568,8 +685,12 @@ class ESP32FirmwareImage(BaseFirmwareImage):
         self.wp_drv = 0
         self.chip_id = 0
         self.min_rev = 0
+        self.min_rev_full = 0
+        self.max_rev_full = 0
+        self.ram_only_header = ram_only_header
 
         self.append_digest = append_digest
+        self.data_length = None
 
         if load_file is not None:
             start = load_file.tell()
@@ -588,6 +709,7 @@ class ESP32FirmwareImage(BaseFirmwareImage):
                 calc_digest = hashlib.sha256()
                 calc_digest.update(load_file.read(end - start))
                 self.calc_digest = calc_digest.digest()  # TODO: decide what to do here?
+                self.data_length = end - start
 
             self.verify()
 
@@ -627,6 +749,27 @@ class ESP32FirmwareImage(BaseFirmwareImage):
                 if not self.is_flash_addr(s.addr)
             ]
 
+            # Patch to support ESP32-C6 union bus memmap
+            # move ".flash.appdesc" segment to the top of the flash segment
+            for segment in flash_segments:
+                if isinstance(segment, ELFSection) and segment.name == ".flash.appdesc":
+                    flash_segments.remove(segment)
+                    flash_segments.insert(0, segment)
+                    break
+
+            # For the bootloader image
+            # move ".dram0.bootdesc" segment to the top of the ram segment
+            # So bootdesc will be at the very top of the binary at 0x20 offset
+            # (in the first segment).
+            for segment in ram_segments:
+                if (
+                    isinstance(segment, ELFSection)
+                    and segment.name == ".dram0.bootdesc"
+                ):
+                    ram_segments.remove(segment)
+                    ram_segments.insert(0, segment)
+                    break
+
             # check for multiple ELF sections that are mapped in the same
             # flash mapping region. This is usually a sign of a broken linker script,
             # but if you have a legitimate use case then let us know
@@ -635,10 +778,9 @@ class ESP32FirmwareImage(BaseFirmwareImage):
                 for segment in flash_segments[1:]:
                     if segment.addr // self.IROM_ALIGN == last_addr // self.IROM_ALIGN:
                         raise FatalError(
-                            "Segment loaded at 0x%08x lands in same 64KB flash mapping "
-                            "as segment loaded at 0x%08x. Can't generate binary. "
+                            f"Segment loaded at {segment.addr:#010x} lands in same {self.IROM_ALIGN // 1024} KB flash mapping "
+                            f"as segment loaded at {last_addr:#010x}. Can't generate binary. "
                             "Suggest changing linker script or ELF to merge sections."
-                            % (segment.addr, last_addr)
                         )
                     last_addr = segment.addr
 
@@ -662,36 +804,76 @@ class ESP32FirmwareImage(BaseFirmwareImage):
                     pad_len += self.IROM_ALIGN
                 return pad_len
 
-            # try to fit each flash segment on a 64kB aligned boundary
-            # by padding with parts of the non-flash segments...
-            while len(flash_segments) > 0:
-                segment = flash_segments[0]
-                pad_len = get_alignment_data_needed(segment)
-                if pad_len > 0:  # need to pad
-                    if len(ram_segments) > 0 and pad_len > self.SEG_HEADER_LEN:
-                        pad_segment = ram_segments[0].split_image(pad_len)
-                        if len(ram_segments[0].data) == 0:
-                            ram_segments.pop(0)
-                    else:
-                        pad_segment = ImageSegment(0, b"\x00" * pad_len, f.tell())
-                    checksum = self.save_segment(f, pad_segment, checksum)
-                    total_segments += 1
-                else:
-                    # write the flash segment
-                    assert (
-                        f.tell() + 8
-                    ) % self.IROM_ALIGN == segment.addr % self.IROM_ALIGN
-                    checksum = self.save_flash_segment(f, segment, checksum)
-                    flash_segments.pop(0)
+            if self.ram_only_header:
+                # write RAM segments first in order to get only RAM segments quantity
+                # and checksum (ROM bootloader will only care for RAM segments and its
+                # correct checksums)
+                for segment in ram_segments:
+                    checksum = self.save_segment(f, segment, checksum, segment.name)
                     total_segments += 1
+                self.append_checksum(f, checksum)
 
-            # flash segments all written, so write any remaining RAM segments
-            for segment in ram_segments:
-                checksum = self.save_segment(f, segment, checksum)
-                total_segments += 1
+                # reversing to match the same section order from linker script
+                flash_segments.reverse()
+                for segment in flash_segments:
+                    pad_len = get_alignment_data_needed(segment)
+                    # Some chips have a non-zero load offset (eg. 0x1000)
+                    # therefore we shift the ROM segments "-load_offset"
+                    # so it will be aligned properly after it is flashed
+                    align_min = (
+                        self.ROM_LOADER.BOOTLOADER_FLASH_OFFSET - self.SEG_HEADER_LEN
+                    )
+                    if pad_len < align_min:
+                        # in case pad_len does not fit minimum alignment,
+                        # pad it to next aligned boundary
+                        pad_len += self.IROM_ALIGN
+
+                    pad_len -= self.ROM_LOADER.BOOTLOADER_FLASH_OFFSET
+                    pad_segment = ImageSegment(0, b"\x00" * pad_len, f.tell())
+                    self.save_segment(f, pad_segment, None, segment.name)
+                    total_segments += 1
+                    # check the alignment
+                    assert (f.tell() + 8 + self.ROM_LOADER.BOOTLOADER_FLASH_OFFSET) % (
+                        self.IROM_ALIGN
+                    ) == segment.addr % self.IROM_ALIGN
+                    # save the flash segment but not saving its checksum neither
+                    # saving the number of flash segments, since ROM bootloader
+                    # should "not see" them
+                    self.save_flash_segment(f, segment)
+                    total_segments += 1
+            else:  # not self.ram_only_header
+                # try to fit each flash segment on a MMU page size aligned boundary
+                # by padding with parts of the non-flash segments...
+                while len(flash_segments) > 0:
+                    segment = flash_segments[0]
+                    pad_len = get_alignment_data_needed(segment)
+                    if pad_len > 0:  # need to pad
+                        if len(ram_segments) > 0 and pad_len > self.SEG_HEADER_LEN:
+                            pad_segment = ram_segments[0].split_image(pad_len)
+                            if len(ram_segments[0].data) == 0:
+                                ram_segments.pop(0)
+                        else:
+                            pad_segment = ImageSegment(0, b"\x00" * pad_len, f.tell())
+                        checksum = self.save_segment(
+                            f, pad_segment, checksum, segment.name
+                        )
+                        total_segments += 1
+                    else:
+                        # write the flash segment
+                        assert (
+                            f.tell() + 8
+                        ) % self.IROM_ALIGN == segment.addr % self.IROM_ALIGN
+                        checksum = self.save_flash_segment(f, segment, checksum)
+                        flash_segments.pop(0)
+                        total_segments += 1
+
+                # flash segments all written, so write any remaining RAM segments
+                for segment in ram_segments:
+                    checksum = self.save_segment(f, segment, checksum, segment.name)
+                    total_segments += 1
 
             if self.secure_pad:
-                # pad the image so that after signing it will end on a a 64KB boundary.
+                # pad the image so that after signing it will end on a a MMU page size boundary.
                 # This ensures all mapped flash content will be verified.
                 if not self.append_digest:
                     raise FatalError(
@@ -710,7 +892,7 @@ class ESP32FirmwareImage(BaseFirmwareImage):
                 elif self.secure_pad == "2":  # Secure Boot V2
                     # after checksum: SHA-256 digest +
                     # signature sector,
-                    # but we place signature sector after the 64KB boundary
+                    # but we place signature sector after the MMU page size boundary
                     space_after_checksum = 32
                 pad_len = (
                     self.IROM_ALIGN - align_past - checksum_space - space_after_checksum
@@ -720,8 +902,9 @@ class ESP32FirmwareImage(BaseFirmwareImage):
                 checksum = self.save_segment(f, pad_segment, checksum)
                 total_segments += 1
 
-            # done writing segments
-            self.append_checksum(f, checksum)
+            if not self.ram_only_header:
+                # done writing segments
+                self.append_checksum(f, checksum)
             image_length = f.tell()
 
             if self.secure_pad:
@@ -730,7 +913,12 @@ class ESP32FirmwareImage(BaseFirmwareImage):
             # kinda hacky: go back to the initial header and write the new segment count
             # that includes padding segments. This header is not checksummed
             f.seek(1)
-            f.write(bytes([total_segments]))
+            if self.ram_only_header:
+                # Update the header with the RAM segments quantity as it should be
+                # visible by the ROM bootloader
+                f.write(bytes([len(ram_segments)]))
+            else:
+                f.write(bytes([total_segments]))
 
             if self.append_digest:
                 # calculate the SHA256 of the whole file and append it
@@ -739,22 +927,15 @@ class ESP32FirmwareImage(BaseFirmwareImage):
                 digest.update(f.read(image_length))
                 f.write(digest.digest())
 
+            if self.pad_to_size:
+                image_length = f.tell()
+                if image_length % self.pad_to_size != 0:
+                    pad_by = self.pad_to_size - (image_length % self.pad_to_size)
+                    f.write(b"\xff" * pad_by)
+
             with open(filename, "wb") as real_file:
                 real_file.write(f.getvalue())
 
-    def save_flash_segment(self, f, segment, checksum=None):
-        """
-        Save the next segment to the image file, return next checksum value if provided
-        """
-        segment_end_pos = f.tell() + len(segment.data) + self.SEG_HEADER_LEN
-        segment_len_remainder = segment_end_pos % self.IROM_ALIGN
-        if segment_len_remainder < 0x24:
-            # Work around a bug in ESP-IDF 2nd stage bootloader, that it didn't map the
-            # last MMU page, if an IROM/DROM segment was < 0x24 bytes
-            # over the page boundary.
-            segment.data += b"\x00" * (0x24 - segment_len_remainder)
-        return self.save_segment(f, segment, checksum)
-
     def load_extended_header(self, load_file):
         def split_byte(n):
             return (n & 0x0F, (n >> 4) & 0x0F)
@@ -781,13 +962,8 @@ class ESP32FirmwareImage(BaseFirmwareImage):
             )
 
         self.min_rev = fields[5]
-
-        # reserved fields in the middle should all be zero
-        if any(f for f in fields[6:-1] if f != 0):
-            print(
-                "Warning: some reserved header fields have non-zero values. "
-                "This image may be from a newer esptool.py?"
-            )
+        self.min_rev_full = fields[6]
+        self.max_rev_full = fields[7]
 
         append_digest = fields[-1]  # last byte is append_digest
         if append_digest in [0, 1]:
@@ -811,8 +987,10 @@ class ESP32FirmwareImage(BaseFirmwareImage):
             join_byte(self.hd_drv, self.wp_drv),
             self.ROM_LOADER.IMAGE_CHIP_ID,
             self.min_rev,
+            self.min_rev_full,
+            self.max_rev_full,
         ]
-        fields += [0] * 8  # padding
+        fields += [0] * 4  # padding
         fields += [append_digest]
 
         packed = struct.pack(self.EXTENDED_HEADER_STRUCT_FMT, *fields)
@@ -855,19 +1033,18 @@ class ESP8266V3FirmwareImage(ESP32FirmwareImage):
                 for segment in flash_segments[1:]:
                     if segment.addr // self.IROM_ALIGN == last_addr // self.IROM_ALIGN:
                         raise FatalError(
-                            "Segment loaded at 0x%08x lands in same 64KB flash mapping "
-                            "as segment loaded at 0x%08x. Can't generate binary. "
+                            f"Segment loaded at {segment.addr:#010x} lands in same {self.IROM_ALIGN // 1024} KB flash mapping "
+                            f"as segment loaded at {last_addr:#010x}. Can't generate binary. "
                             "Suggest changing linker script or ELF to merge sections."
-                            % (segment.addr, last_addr)
                         )
                     last_addr = segment.addr
 
-            # try to fit each flash segment on a 64kB aligned boundary
+            # try to fit each flash segment on a MMU page size aligned boundary
             # by padding with parts of the non-flash segments...
             while len(flash_segments) > 0:
                 segment = flash_segments[0]
                 # remove 8 bytes empty data for insert segment header
-                if segment.name == ".flash.rodata":
+                if isinstance(segment, ELFSection) and segment.name == ".flash.rodata":
                     segment.data = segment.data[8:]
                 # write the flash segment
                 checksum = self.save_segment(f, segment, checksum)
@@ -999,24 +1176,86 @@ class ESP32C2FirmwareImage(ESP32FirmwareImage):
     """ESP32C2 Firmware Image almost exactly the same as ESP32FirmwareImage"""
 
     ROM_LOADER = ESP32C2ROM
-
-    def set_mmu_page_size(self, size):
-        if size not in [16384, 32768, 65536]:
-            raise FatalError("{} is not a valid page size.".format(size))
-        self.IROM_ALIGN = size
+    MMU_PAGE_SIZE_CONF = (16384, 32768, 65536)
 
 
 ESP32C2ROM.BOOTLOADER_IMAGE = ESP32C2FirmwareImage
 
 
+class ESP32C6FirmwareImage(ESP32FirmwareImage):
+    """ESP32C6 Firmware Image almost exactly the same as ESP32FirmwareImage"""
+
+    ROM_LOADER: Type[ESPLoader] = ESP32C6ROM
+    MMU_PAGE_SIZE_CONF = (8192, 16384, 32768, 65536)
+
+
+ESP32C6ROM.BOOTLOADER_IMAGE = ESP32C6FirmwareImage
+
+
+class ESP32C61FirmwareImage(ESP32C6FirmwareImage):
+    """ESP32C61 Firmware Image almost exactly the same as ESP32C6FirmwareImage"""
+
+    ROM_LOADER = ESP32C61ROM
+
+
+ESP32C61ROM.BOOTLOADER_IMAGE = ESP32C61FirmwareImage
+
+
+class ESP32C5FirmwareImage(ESP32FirmwareImage):
+    """ESP32C5 Firmware Image almost exactly the same as ESP32FirmwareImage"""
+
+    ROM_LOADER = ESP32C5ROM
+
+
+ESP32C5ROM.BOOTLOADER_IMAGE = ESP32C5FirmwareImage
+
+
+class ESP32C5BETA3FirmwareImage(ESP32C6FirmwareImage):
+    """ESP32C5BETA3 Firmware Image almost exactly the same as ESP32C6FirmwareImage"""
+
+    ROM_LOADER = ESP32C5BETA3ROM
+
+
+ESP32C5BETA3ROM.BOOTLOADER_IMAGE = ESP32C5BETA3FirmwareImage
+
+
+class ESP32H4FirmwareImage(ESP32C6FirmwareImage):
+    """ESP32H4 Firmware Image almost exactly the same as ESP32FirmwareImage"""
+
+    ROM_LOADER = ESP32H4ROM
+
+
+ESP32H4ROM.BOOTLOADER_IMAGE = ESP32H4FirmwareImage
+
+
+class ESP32P4FirmwareImage(ESP32FirmwareImage):
+    """ESP32P4 Firmware Image almost exactly the same as ESP32FirmwareImage"""
+
+    ROM_LOADER = ESP32P4ROM
+
+
+ESP32P4ROM.BOOTLOADER_IMAGE = ESP32P4FirmwareImage
+
+
+class ESP32H2FirmwareImage(ESP32C6FirmwareImage):
+    """ESP32H2 Firmware Image almost exactly the same as ESP32FirmwareImage"""
+
+    ROM_LOADER = ESP32H2ROM
+
+
+ESP32H2ROM.BOOTLOADER_IMAGE = ESP32H2FirmwareImage
+
+
+class ESP32H21FirmwareImage(ESP32C6FirmwareImage):
+    """ESP32H21 Firmware Image almost exactly the same as ESP32FirmwareImage"""
+
+    ROM_LOADER = ESP32H21ROM
+
+
+ESP32H21ROM.BOOTLOADER_IMAGE = ESP32H21FirmwareImage
+
+
 class ELFFile(object):
-    SEC_TYPE_PROGBITS = 0x01
-    SEC_TYPE_STRTAB = 0x03
-    SEC_TYPE_INITARRAY = 0x0E
-    SEC_TYPE_FINIARRAY = 0x0F
-
-    PROG_SEC_TYPES = (SEC_TYPE_PROGBITS, SEC_TYPE_INITARRAY, SEC_TYPE_FINIARRAY)
-
     LEN_SEC_HEADER = 0x28
 
     SEG_TYPE_LOAD = 0x01
@@ -1077,6 +1316,22 @@ class ELFFile(object):
         self._read_segments(f, _phoff, _phnum, shstrndx)
 
     def _read_sections(self, f, section_header_offs, section_header_count, shstrndx):
+        SEC_TYPE_PROGBITS = 0x01
+        SEC_TYPE_STRTAB = 0x03
+        SEC_TYPE_NOBITS = 0x08  # e.g. .bss section
+        SEC_TYPE_INITARRAY = 0x0E
+        SEC_TYPE_FINIARRAY = 0x0F
+        SEC_TYPE_PREINITARRAY = 0x10
+
+        PROG_SEC_TYPES = (
+            SEC_TYPE_PROGBITS,
+            SEC_TYPE_INITARRAY,
+            SEC_TYPE_FINIARRAY,
+            SEC_TYPE_PREINITARRAY,
+        )
+
+        KNOWN_SEC_TYPES = PROG_SEC_TYPES + (SEC_TYPE_NOBITS, SEC_TYPE_STRTAB)
+
         f.seek(section_header_offs)
         len_bytes = section_header_count * self.LEN_SEC_HEADER
         section_header = f.read(len_bytes)
@@ -1095,21 +1350,23 @@ class ELFFile(object):
         section_header_offsets = range(0, len(section_header), self.LEN_SEC_HEADER)
 
         def read_section_header(offs):
-            name_offs, sec_type, _flags, lma, sec_offs, size = struct.unpack_from(
-                "<LLLLLL", section_header[offs:]
-            )
+            (
+                name_offs,
+                sec_type,
+                _flags,
+                lma,
+                sec_offs,
+                size,
+            ) = struct.unpack_from("<LLLLLL", section_header[offs:])
             return (name_offs, sec_type, lma, size, sec_offs)
 
-        all_sections = [read_section_header(offs) for offs in section_header_offsets]
-        prog_sections = [s for s in all_sections if s[1] in ELFFile.PROG_SEC_TYPES]
-
         # search for the string table section
-        if not (shstrndx * self.LEN_SEC_HEADER) in section_header_offsets:
+        if (shstrndx * self.LEN_SEC_HEADER) not in section_header_offsets:
             raise FatalError("ELF file has no STRTAB section at shstrndx %d" % shstrndx)
         _, sec_type, _, sec_size, sec_offs = read_section_header(
             shstrndx * self.LEN_SEC_HEADER
         )
-        if sec_type != ELFFile.SEC_TYPE_STRTAB:
+        if sec_type != SEC_TYPE_STRTAB:
             print(
                 "WARNING: ELF file has incorrect STRTAB section type 0x%02x" % sec_type
             )
@@ -1127,12 +1384,38 @@ class ELFFile(object):
             f.seek(offs)
             return f.read(size)
 
-        prog_sections = [
-            ELFSection(lookup_string(n_offs), lma, read_data(offs, size))
-            for (n_offs, _type, lma, size, offs) in prog_sections
-            if lma != 0 and size > 0
-        ]
-        self.sections = prog_sections
+        all_sections = [read_section_header(offs) for offs in section_header_offsets]
+
+        self.sections = []
+        self.nobits_sections = []
+        # Process all sections and raise an error if an unknown section type is found
+        for section in all_sections:
+            n_offs, sec_type, lma, size, offs = section
+
+            # Skip sections with lma == 0 or size == 0
+            if lma == 0 or size == 0:
+                continue
+
+            if sec_type not in KNOWN_SEC_TYPES:
+                print(f"WARNING: Unknown section type {sec_type:#04x} in ELF file")
+                continue
+
+            if sec_type in PROG_SEC_TYPES:
+                self.sections.append(
+                    ELFSection(
+                        lookup_string(n_offs),
+                        lma,
+                        read_data(offs, size),
+                    )
+                )
+            elif sec_type == SEC_TYPE_NOBITS:
+                self.nobits_sections.append(
+                    ELFSection(
+                        lookup_string(n_offs),
+                        lma,
+                        b"",
+                    )
+                )
 
     def _read_segments(self, f, segment_header_offs, segment_header_count, shstrndx):
         f.seek(segment_header_offs)
