@@ -13,6 +13,8 @@ import string
 import struct
 import sys
 import time
+from typing import Dict, Optional
+
 
 from .config import load_config_file
 from .reset import (
@@ -59,7 +61,7 @@ except ImportError:
     print(
         "The installed version (%s) of pyserial appears to be too old for esptool.py "
         "(Python interpreter %s). Check the README for installation instructions."
-        % (sys.VERSION, sys.executable)
+        % (serial.VERSION, sys.executable)
     )
     raise
 except Exception:
@@ -96,14 +98,8 @@ DEFAULT_SERIAL_WRITE_TIMEOUT = cfg.getfloat("serial_write_timeout", 10)
 DEFAULT_CONNECT_ATTEMPTS = cfg.getint("connect_attempts", 7)
 # Number of times to try writing a data block
 WRITE_BLOCK_ATTEMPTS = cfg.getint("write_block_attempts", 3)
-
-STUBS_DIR = os.path.join(os.path.dirname(__file__), "./targets/stub_flasher/")
-
-
-def get_stub_json_path(chip_name):
-    chip_name = strip_chip_name(chip_name)
-    chip_name = chip_name.replace("esp", "")
-    return STUBS_DIR + "stub_flasher_" + chip_name + ".json"
+# Number of times to try opening the serial port
+DEFAULT_OPEN_PORT_ATTEMPTS = cfg.getint("open_port_attempts", 1)
 
 
 def timeout_per_mb(seconds_per_mb, size_bytes):
@@ -147,16 +143,13 @@ def stub_and_esp32_function_only(func):
     )
 
 
-def esp32s3_or_newer_function_only(func):
-    """Attribute for a function only supported by ESP32S3 and later chips ROM"""
-    return check_supported_function(
-        func, lambda o: o.CHIP_NAME not in ["ESP8266", "ESP32", "ESP32-S2"]
-    )
-
-
 class StubFlasher:
-    def __init__(self, json_path):
-        with open(json_path) as json_file:
+    STUB_DIR = os.path.join(os.path.dirname(__file__), "targets", "stub_flasher")
+    # directories will be searched in the order of STUB_SUBDIRS
+    STUB_SUBDIRS = ["1", "2"]
+
+    def __init__(self, chip_name):
+        with open(self.get_json_path(chip_name)) as json_file:
             stub = json.load(json_file)
 
         self.text = base64.b64decode(stub["text"])
@@ -170,6 +163,28 @@ class StubFlasher:
             self.data = None
             self.data_start = None
 
+        self.bss_start = stub.get("bss_start")
+
+    def get_json_path(self, chip_name):
+        chip_name = strip_chip_name(chip_name)
+        for i, subdir in enumerate(self.STUB_SUBDIRS):
+            json_path = os.path.join(self.STUB_DIR, subdir, f"{chip_name}.json")
+            if os.path.exists(json_path):
+                if i:
+                    print(
+                        f"Warning: Stub version {self.STUB_SUBDIRS[0]} doesn't exist, using {subdir} instead"
+                    )
+
+                return json_path
+        else:
+            raise FileNotFoundError(f"Stub flasher JSON file for {chip_name} not found")
+
+    @classmethod
+    def set_preferred_stub_subdir(cls, subdir):
+        if subdir in cls.STUB_SUBDIRS:
+            cls.STUB_SUBDIRS.remove(subdir)
+            cls.STUB_SUBDIRS.insert(0, subdir)
+
 
 class ESPLoader(object):
     """Base class providing access to ESP ROM & software stub bootloaders.
@@ -177,14 +192,15 @@ class ESPLoader(object):
 
     Don't instantiate this base class directly, either instantiate a subclass or
     call cmds.detect_chip() which will interrogate the chip and return the
-    appropriate subclass instance.
+    appropriate subclass instance. You can also use a context manager as
+    "with detect_chip() as esp:" to ensure the serial port is closed when done.
 
     """
 
     CHIP_NAME = "Espressif device"
     IS_STUB = False
-
-    FPGA_SLOW_BOOT = False
+    STUB_CLASS: Optional[object] = None
+    BOOTLOADER_IMAGE: Optional[object] = None
 
     DEFAULT_PORT = "/dev/ttyUSB0"
 
@@ -201,7 +217,7 @@ class ESPLoader(object):
     ESP_WRITE_REG = 0x09
     ESP_READ_REG = 0x0A
 
-    # Some comands supported by ESP32 and later chips ROM bootloader (or -8266 w/ stub)
+    # Some commands supported by ESP32 and later chips ROM bootloader (or -8266 w/ stub)
     ESP_SPI_SET_PARAMS = 0x0B
     ESP_SPI_ATTACH = 0x0D
     ESP_READ_FLASH_SLOW = 0x0E  # ROM only, much slower than the stub flash read
@@ -245,6 +261,9 @@ class ESPLoader(object):
 
     UART_DATE_REG_ADDR = 0x60000078
 
+    # Whether the SPI peripheral sends from MSB of 32-bit register, or the MSB of valid LSB bits.
+    SPI_ADDR_REG_MSB = True
+
     # This ROM address has a different value on each chip model
     CHIP_DETECT_MAGIC_REG_ADDR = 0x40001000
 
@@ -254,9 +273,6 @@ class ESPLoader(object):
     IROM_MAP_START = 0x40200000
     IROM_MAP_END = 0x40300000
 
-    # The number of bytes in the UART response that signify command status
-    STATUS_BYTES_LENGTH = 2
-
     # Bootloader flashing offset
     BOOTLOADER_FLASH_OFFSET = 0x0
 
@@ -273,11 +289,22 @@ class ESPLoader(object):
     # Chip IDs that are no longer supported by esptool
     UNSUPPORTED_CHIPS = {6: "ESP32-S3(beta 3)"}
 
+    # Number of attempts to write flash data
+    WRITE_FLASH_ATTEMPTS = 2
+
+    FLASH_ENCRYPTED_WRITE_ALIGN = 16
+    KEY_PURPOSES: Dict[int, str] = {}
+    EFUSE_MAX_KEY = 5
+
+    # Chip uses magic number for chip type autodetection
+    USES_MAGIC_VALUE = True
+
     def __init__(self, port=DEFAULT_PORT, baud=ESP_ROM_BAUD, trace_enabled=False):
         """Base constructor for ESPLoader bootloader interaction
 
         Don't call this constructor, either instantiate a specific
-        ROM class directly, or use cmds.detect_chip().
+        ROM class directly, or use cmds.detect_chip(). You can use the with
+        statement to ensure the serial port is closed when done.
 
         This base class has all of the instance methods for bootloader
         functionality supported across various chips & stub
@@ -293,15 +320,54 @@ class ESPLoader(object):
         # Device-and-runtime-specific cache
         self.cache = {
             "flash_id": None,
-            "chip_id": None,
             "uart_no": None,
+            "usb_pid": None,
+            "security_info": None,
         }
 
         if isinstance(port, str):
             try:
-                self._port = serial.serial_for_url(port)
-            except serial.serialutil.SerialException:
-                raise FatalError(f"Could not open {port}, the port doesn't exist")
+                self._port = serial.serial_for_url(
+                    port, exclusive=True, do_not_open=True
+                )
+                if sys.platform == "win32":
+                    # When opening a port on Windows,
+                    # the RTS/DTR (active low) lines
+                    # need to be set to False (pulled high)
+                    # to avoid unwanted chip reset
+                    self._port.rts = False
+                    self._port.dtr = False
+                self._port.open()
+            except serial.serialutil.SerialException as e:
+                port_issues = [
+                    [  # does not exist error
+                        re.compile(r"Errno 2|FileNotFoundError", re.IGNORECASE),
+                        "Check if the port is correct and ESP connected",
+                    ],
+                    [  # busy port error
+                        re.compile(r"Access is denied", re.IGNORECASE),
+                        "Check if the port is not used by another task",
+                    ],
+                ]
+                if sys.platform.startswith("linux"):
+                    port_issues.append(
+                        [  # permission denied error
+                            re.compile(r"Permission denied", re.IGNORECASE),
+                            ("Try to add user into dialout or uucp group."),
+                        ],
+                    )
+
+                hint_msg = ""
+                for port_issue in port_issues:
+                    if port_issue[0].search(str(e)):
+                        hint_msg = f"\nHint: {port_issue[1]}\n"
+                        break
+
+                raise FatalError(
+                    f"Could not open {port}, the port is busy or doesn't exist."
+                    f"\n({e})\n"
+                    f"{hint_msg}"
+                )
         else:
             self._port = port
         self._slip_reader = slip_reader(self._port, self.trace)
@@ -319,6 +385,12 @@ class ESPLoader(object):
             # need to set the property back to None or it will continue to fail
             self._port.write_timeout = None
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        self._port.close()
+
     @property
     def serial_port(self):
         return self._port.port
@@ -412,9 +484,35 @@ class ESPLoader(object):
                 if op is None or op_ret == op:
                     return val, data
                 if byte(data, 0) != 0 and byte(data, 1) == self.ROM_INVALID_RECV_MSG:
-                    # Unsupported read_reg can result in
+
+                    def drain_input_buffer(buffering_time=0.2):
+                        """
+                        Actively drain the input buffer by reading data
+                        for a specified time. Simple approach for some
+                        drivers that have issues with the buffer flushing.
+
+                        Args:
+                            buffering_time: Time in seconds to wait for
+                            the buffer to fill.
+                        """
+                        time.sleep(buffering_time)
+                        original_timeout = self._port.timeout
+                        # Set a very short timeout for draining
+                        self._port.timeout = 0.001
+
+                        # Unsupported command response is sent 8 times and has
+                        # 14 bytes length including delimiter 0xC0 bytes.
+                        # At least part of it is read as a command response,
+                        # but to be safe, read it all.
+                        self._port.read(14 * 8)
+
+                        # Restore original timeout
+                        self._port.timeout = original_timeout
+                        self.flush_input()
+
+                    # Unsupported command can result in
                     # more than one error response for some reason
-                    self.flush_input()
+                    drain_input_buffer(0.2)
                     raise UnsupportedCommandError(self, op)
 
         finally:
@@ -424,7 +522,13 @@ class ESPLoader(object):
         raise FatalError("Response doesn't match request")
 
     def check_command(
-        self, op_description, op=None, data=b"", chk=0, timeout=DEFAULT_TIMEOUT
+        self,
+        op_description,
+        op=None,
+        data=b"",
+        chk=0,
+        resp_data_len=0,
+        timeout=DEFAULT_TIMEOUT,
     ):
         """
         Execute a command with 'command', check the result code and throw an appropriate
@@ -432,28 +536,40 @@ class ESPLoader(object):
 
         Returns the "result" of a successful command.
         """
+
+        # The status bytes are first two bytes after the expected response data,
+        # ROM bootloaders of the chips (except ESP8266) return another 2 bytes,
+        # which are reserved for future use, these are ignored here.
+        STATUS_BYTES_LENGTH = 2
+
+        # Execute the command and get the result
         val, data = self.command(op, data, chk, timeout=timeout)
 
-        # things are a bit weird here, bear with us
+        # Check if we have enough data,
+        # including the expected response data and status bytes
+        if len(data) < resp_data_len + STATUS_BYTES_LENGTH:
+            # If we don't have enough data, check the first 2 bytes as status
+            status_bytes = data[0:2]
+            # Only care if the first byte is non-zero.
+            # If it is, the second byte is a reason.
+            if status_bytes[0] != 0:
+                raise FatalError.WithResult(f"Failed to {op_description}", status_bytes)
+            else:
+                raise FatalError(
+                    f"Failed to {op_description}. "
+                    f"Only got {len(data)} byte status response."
+                )
+        # The status bytes are positioned after the expected response data
+        # (first two bytes after resp_data_len are the status bytes)
+        status_bytes = data[resp_data_len : resp_data_len + STATUS_BYTES_LENGTH]
+        # Only care if the first byte is non-zero.
+        # If it is, the second byte is a reason.
+        if status_bytes[0] != 0:
+            raise FatalError.WithResult(f"Failed to {op_description}", status_bytes)
 
-        # the status bytes are the last 2/4 bytes in the data (depending on chip)
-        if len(data) < self.STATUS_BYTES_LENGTH:
-            raise FatalError(
-                "Failed to %s. Only got %d byte status response."
-                % (op_description, len(data))
-            )
-        status_bytes = data[-self.STATUS_BYTES_LENGTH :]
-        # only care if the first one is non-zero. If it is, the second byte is a reason.
-        if byte(status_bytes, 0) != 0:
-            raise FatalError.WithResult("Failed to %s" % op_description, status_bytes)
-
-        # if we had more data than just the status bytes, return it as the result
-        # (this is used by the md5sum command, maybe other commands?)
-        if len(data) > self.STATUS_BYTES_LENGTH:
-            return data[: -self.STATUS_BYTES_LENGTH]
+        if resp_data_len > 0:
+            return data[:resp_data_len]
         else:
-            # otherwise, just return the 'val' field which comes from the reply header
-            # (this is used by read_reg)
             return val
 
     def flush_input(self):
@@ -467,7 +583,7 @@ class ESPLoader(object):
 
         # ROM bootloaders send some non-zero "val" response. The flasher stub sends 0.
         # If we receive 0 then it probably indicates that the chip wasn't or couldn't be
-        # reseted properly and esptool is talking to the flasher stub.
+        # reset properly and esptool is talking to the flasher stub.
         self.sync_stub_detected = val == 0
 
         for _ in range(7):
@@ -475,6 +591,9 @@ class ESPLoader(object):
             self.sync_stub_detected &= val == 0
 
     def _get_pid(self):
+        if self.cache["usb_pid"] is not None:
+            return self.cache["usb_pid"]
+
         if list_ports is None:
             print(
                 "\nListing all serial ports is currently not available. "
@@ -494,16 +613,19 @@ class ESPLoader(object):
         if active_port.startswith("/dev/") and os.path.islink(active_port):
             active_port = os.path.realpath(active_port)
 
+        active_ports = [active_port]
+
         # The "cu" (call-up) device has to be used for outgoing communication on MacOS
         if sys.platform == "darwin" and "tty" in active_port:
-            active_port = [active_port, active_port.replace("tty", "cu")]
+            active_ports.append(active_port.replace("tty", "cu"))
         ports = list_ports.comports()
         for p in ports:
-            if p.device in active_port:
+            if p.device in active_ports:
+                self.cache["usb_pid"] = p.pid
                 return p.pid
         print(
-            "\nFailed to get PID of a device on {}, "
-            "using standard reset sequence.".format(active_port)
+            f"\nFailed to get PID of a device on {active_port}, "
+            "using standard reset sequence."
         )
 
     def _connect_attempt(self, reset_strategy, mode="default_reset"):
@@ -590,7 +712,7 @@ class ESPLoader(object):
 
         # This FPGA delay is for Espressif internal use
         if (
-            self.FPGA_SLOW_BOOT
+            self.CHIP_NAME == "ESP32"
             and os.environ.get("ESPTOOL_ENV_FPGA", "").strip() == "1"
         ):
             delay = extra_delay = 7
@@ -627,6 +749,15 @@ class ESPLoader(object):
                 "Connection may fail if the chip is not in bootloader "
                 "or flasher stub mode.",
             )
+
+        if self._port.name.startswith("socket:"):
+            mode = "no_reset"  # not possible to toggle DTR/RTS over a TCP socket
+            print(
+                "Note: It's not possible to reset the chip over a TCP socket. "
+                "Automatic resetting to bootloader has been disabled, "
+                "reset the chip manually."
+            )
+
         print("Connecting...", end="")
         sys.stdout.flush()
         last_error = None
@@ -644,8 +775,17 @@ class ESPLoader(object):
             print("")  # end 'Connecting...' line
 
         if last_error is not None:
+            additional_msg = ""
+            if self.CHIP_NAME == "ESP32-C2" and self._port.baudrate < 115200:
+                additional_msg = (
+                    "\nNote: Please set a higher baud rate (--baud)"
+                    " if ESP32-C2 doesn't connect"
+                    " (at least 115200 Bd is recommended)."
+                )
+            self._port.close()
             raise FatalError(
                 "Failed to connect to {}: {}"
+                f"{additional_msg}"
                 "\nFor troubleshooting steps visit: "
                 "https://docs.espressif.com/projects/esptool/en/latest/troubleshooting.html".format(  # noqa E501
                     self.CHIP_NAME, last_error
@@ -653,41 +793,80 @@ class ESPLoader(object):
             )
 
         if not detecting:
-            try:
-                from .targets import ROM_LIST
-
-                # check the date code registers match what we expect to see
-                chip_magic_value = self.read_reg(ESPLoader.CHIP_DETECT_MAGIC_REG_ADDR)
-                if chip_magic_value not in self.CHIP_DETECT_MAGIC_VALUE:
-                    actually = None
-                    for cls in ROM_LIST:
-                        if chip_magic_value in cls.CHIP_DETECT_MAGIC_VALUE:
-                            actually = cls
-                            break
-                    if warnings and actually is None:
-                        print(
-                            "WARNING: This chip doesn't appear to be a %s "
-                            "(chip magic value 0x%08x). "
-                            "Probably it is unsupported by this version of esptool."
-                            % (self.CHIP_NAME, chip_magic_value)
-                        )
-                    else:
-                        raise FatalError(
-                            "This chip is %s not %s. Wrong --chip argument?"
-                            % (actually.CHIP_NAME, self.CHIP_NAME)
-                        )
-            except UnsupportedCommandError:
-                self.secure_download_mode = True
+            from .targets import ROM_LIST
 
+            # Check if chip supports reading chip ID from the get-security-info command
             try:
-                self.check_chip_id()
-            except UnsupportedCommandError:
-                # Fix for ROM not responding in SDM, reconnect and try again
-                if self.secure_download_mode:
-                    self._connect_attempt(mode, reset_sequence[0])
-                    self.check_chip_id()
+                # get_chip_id() raises FatalError if the chip does not have a chip ID
+                # (ESP32-S2)
+                chip_id = self.get_chip_id()
+                si = self.get_security_info()
+                self.secure_download_mode = si["parsed_flags"]["SECURE_DOWNLOAD_ENABLE"]
+            except (UnsupportedCommandError, FatalError):
+                chip_id = None
+                # Try to read the chip magic value to verify the chip type
+                # (ESP8266, ESP32, ESP32-S2)
+                try:
+                    chip_magic_value = self.read_reg(
+                        ESPLoader.CHIP_DETECT_MAGIC_REG_ADDR
+                    )
+                except UnsupportedCommandError:
+                    # If the chip does not support reading the chip magic value,
+                    # it is ESP32-S2 in SDM
+                    chip_magic_value = None
+                    self.secure_download_mode = True
+
+            detected = None
+            chip_arg_wrong = False
+
+            # If we can read chip ID (ESP32-S3 and later), verify the ID
+            if chip_id and (self.USES_MAGIC_VALUE or chip_id != self.IMAGE_CHIP_ID):
+                chip_arg_wrong = True
+                for cls in ROM_LIST:
+                    if not cls.USES_MAGIC_VALUE and chip_id == cls.IMAGE_CHIP_ID:
+                        detected = cls
+                        break
+            # If we can't read chip ID (ESP8266, ESP32, ESP32-S2),
+            # try to verify the chip by magic value
+            elif (
+                not chip_id
+                and not self.secure_download_mode
+                and (not self.USES_MAGIC_VALUE or chip_magic_value != self.MAGIC_VALUE)
+            ):
+                chip_arg_wrong = True
+                for cls in ROM_LIST:
+                    if cls.USES_MAGIC_VALUE and chip_magic_value == cls.MAGIC_VALUE:
+                        detected = cls
+                        break
+            # If we can't read chip ID and the chip is in SDM, it is ESP32-S2
+            elif (
+                not chip_id
+                and self.secure_download_mode
+                and self.CHIP_NAME != "ESP32-S2"
+            ):
+                chip_arg_wrong = True
+                detected = "ESP32-S2"
+
+            if chip_arg_wrong:
+                if warnings and detected is None:
+                    specifier = (
+                        f"(read chip ID {chip_id})"
+                        if chip_id
+                        else f"(read chip magic value {chip_magic_value:#08x})"
+                    )
+                    print(
+                        f"WARNING: This chip doesn't appear to be an {self.CHIP_NAME} "
+                        f"{specifier}. Probably it is unsupported by this version "
+                        "of esptool. Will attempt to continue anyway."
+                    )
                 else:
-                    raise
+                    chip_type = (
+                        detected if isinstance(detected, str) else detected.CHIP_NAME
+                    )
+                    raise FatalError(
+                        f"This chip is {chip_type}, not {self.CHIP_NAME}. "
+                        "Wrong --chip argument?"
+                    )
             self._post_connect()
 
     def _post_connect(self):
@@ -699,17 +878,10 @@ class ESPLoader(object):
 
     def read_reg(self, addr, timeout=DEFAULT_TIMEOUT):
         """Read memory address in target"""
-        # we don't call check_command here because read_reg() function is called
-        # when detecting chip type, and the way we check for success
-        # (STATUS_BYTES_LENGTH) is different for different chip types (!)
-        val, data = self.command(
-            self.ESP_READ_REG, struct.pack("<I", addr), timeout=timeout
+        command = struct.pack("<I", addr)
+        return self.check_command(
+            "read target memory", self.ESP_READ_REG, command, timeout=timeout
         )
-        if byte(data, 0) != 0:
-            raise FatalError.WithResult(
-                "Failed to read register address %08x" % addr, data
-            )
-        return val
 
     def write_reg(self, addr, value, mask=0xFFFFFFFF, delay_us=0, delay_after_us=0):
         """Write to memory address in target"""
@@ -741,20 +913,23 @@ class ESPLoader(object):
         """Start downloading an application image to RAM"""
         # check we're not going to overwrite a running stub with this data
         if self.IS_STUB:
-            stub = StubFlasher(get_stub_json_path(self.CHIP_NAME))
+            stub = StubFlasher(self.CHIP_NAME)
             load_start = offset
             load_end = offset + size
-            for start, end in [
-                (stub.data_start, stub.data_start + len(stub.data)),
-                (stub.text_start, stub.text_start + len(stub.text)),
+            for stub_start, stub_end in [
+                (
+                    stub.bss_start or stub.data_start,
+                    stub.data_start + len(stub.data),
+                ),  # DRAM = bss+data
+                (stub.text_start, stub.text_start + len(stub.text)),  # IRAM
             ]:
-                if load_start < end and load_end > start:
+                if load_start < stub_end and load_end > stub_start:
                     raise FatalError(
                         "Software loader is resident at 0x%08x-0x%08x. "
                         "Can't load binary at overlapping address range 0x%08x-0x%08x. "
                         "Either change binary loading address, or use the --no-stub "
                         "option to disable the software loader."
-                        % (start, end, load_start, load_end)
+                        % (stub_start, stub_end, load_start, load_end)
                     )
 
         return self.check_command(
@@ -789,7 +964,7 @@ class ESPLoader(object):
                 raise
             pass
 
-    def flash_begin(self, size, offset, begin_rom_encrypted=False):
+    def flash_begin(self, size, offset, begin_rom_encrypted=False, logging=True):
         """
         Start downloading to Flash (performs an erase)
 
@@ -814,7 +989,7 @@ class ESPLoader(object):
         self.check_command(
             "enter Flash download mode", self.ESP_FLASH_BEGIN, params, timeout=timeout
         )
-        if size != 0 and not self.IS_STUB:
+        if size != 0 and not self.IS_STUB and logging:
             print("Took %.2fs to erase flash block" % (time.time() - t))
         return num_blocks
 
@@ -888,37 +1063,123 @@ class ESPLoader(object):
         """Read flash type bit field from eFuse. Returns 0, 1, None (not present)"""
         return None  # not implemented for all chip targets
 
-    def get_security_info(self):
-        res = self.check_command("get security info", self.ESP_GET_SECURITY_INFO, b"")
-        esp32s2 = True if len(res) == 12 else False
-        res = struct.unpack("<IBBBBBBBB" if esp32s2 else "<IBBBBBBBBII", res)
-        return {
+    def get_security_info(self, cache=True):
+        """
+        Get security information from the ESP device including flags,
+        flash encryption count,
+        key purposes, chip ID, and API version.
+
+        The security info command response format according to the ESP32-S3
+        documentation:
+        - 32 bits flags
+        - 1 byte flash_crypt_cnt
+        - 7x1 byte key_purposes
+        - 32-bit word chip_id (ESP32-S3 and later)
+        - 32-bit word eco_version/api_version (ESP32-S3 and later)
+
+        Returns a dictionary with parsed security information and individual
+        flag status.
+        """
+        if cache and self.cache["security_info"] is not None:
+            return self.cache["security_info"]
+
+        # The following mapping was taken from the ROM code
+        # This mapping is same across all targets in the ROM
+        SECURITY_INFO_FLAG_MAP = {
+            "SECURE_BOOT_EN": (1 << 0),
+            "SECURE_BOOT_AGGRESSIVE_REVOKE": (1 << 1),
+            "SECURE_DOWNLOAD_ENABLE": (1 << 2),
+            "SECURE_BOOT_KEY_REVOKE0": (1 << 3),
+            "SECURE_BOOT_KEY_REVOKE1": (1 << 4),
+            "SECURE_BOOT_KEY_REVOKE2": (1 << 5),
+            "SOFT_DIS_JTAG": (1 << 6),
+            "HARD_DIS_JTAG": (1 << 7),
+            "DIS_USB": (1 << 8),
+            "DIS_DOWNLOAD_DCACHE": (1 << 9),
+            "DIS_DOWNLOAD_ICACHE": (1 << 10),
+        }
+
+        def parse_security_flags(flags_value):
+            """Parse security flags into individual boolean values"""
+            parsed_flags = {}
+            for flag_name, flag_mask in SECURITY_INFO_FLAG_MAP.items():
+                parsed_flags[flag_name] = (flags_value & flag_mask) != 0
+            return parsed_flags
+
+        # Try with 20 bytes first (most chips), fall back to 12 bytes (ESP32-S2)
+        try:
+            res = self.check_command(
+                "get security info",
+                self.ESP_GET_SECURITY_INFO,
+                b"",
+                resp_data_len=20,
+            )
+            res = struct.unpack("<IBBBBBBBBII", res)
+            esp32s2 = False
+        except FatalError:
+            res = self.check_command(
+                "get security info",
+                self.ESP_GET_SECURITY_INFO,
+                b"",
+                resp_data_len=12,
+            )
+            res = struct.unpack("<IBBBBBBBB", res)
+            esp32s2 = True
+
+        security_info = {
             "flags": res[0],
             "flash_crypt_cnt": res[1],
             "key_purposes": res[2:9],
             "chip_id": None if esp32s2 else res[9],
             "api_version": None if esp32s2 else res[10],
+            "parsed_flags": parse_security_flags(res[0]),
         }
 
-    @esp32s3_or_newer_function_only
+        self.cache["security_info"] = security_info
+        return security_info
+
     def get_chip_id(self):
-        if self.cache["chip_id"] is None:
-            res = self.check_command(
-                "get security info", self.ESP_GET_SECURITY_INFO, b""
+        chip_id = self.get_security_info()["chip_id"]
+        if chip_id is None:
+            raise FatalError(
+                "Security info command does not contain chip ID. "
+                "This is expected for ESP32-S2 which doesn't support chip ID "
+                "in security info."
             )
-            res = struct.unpack(
-                "<IBBBBBBBBI", res[:16]
-            )  # 4b flags, 1b flash_crypt_cnt, 7*1b key_purposes, 4b chip_id
-            self.cache["chip_id"] = res[9]  # 2/4 status bytes invariant
-        return self.cache["chip_id"]
+        return chip_id
 
     def get_uart_no(self):
         """
         Read the UARTDEV_BUF_NO register to get the number of the currently used console
         """
-        if self.cache["uart_no"] is None:
-            self.cache["uart_no"] = self.read_reg(self.UARTDEV_BUF_NO) & 0xFF
-        return self.cache["uart_no"]
+        # Some ESP chips do not have this register
+        try:
+            if self.cache["uart_no"] is None:
+                self.cache["uart_no"] = self.read_reg(self.UARTDEV_BUF_NO) & 0xFF
+            return self.cache["uart_no"]
+        except AttributeError:
+            return None
+
+    def uses_usb_jtag_serial(self):
+        """
+        Check if the chip uses USB JTAG/SERIAL mode.
+        """
+        return False
+
+    def uses_usb_otg(self):
+        """
+        Check if the chip uses USB OTG mode.
+        """
+        return False
+
+    def get_usb_mode(self):
+        """
+        Get the USB mode of the chip: USB-Serial/JTAG or USB-OTG. If the usb_mode is None, external USB-UART is used.
+        """
+        usb_jtag_serial = self.uses_usb_jtag_serial()
+        usb_otg = self.uses_usb_otg()
+
+        return "USB-Serial/JTAG" if usb_jtag_serial else "USB-OTG" if usb_otg else None
 
     @classmethod
     def parse_flash_size_arg(cls, arg):
@@ -946,12 +1207,16 @@ class ESPLoader(object):
 
     def run_stub(self, stub=None):
         if stub is None:
-            stub = StubFlasher(get_stub_json_path(self.CHIP_NAME))
+            stub = StubFlasher(self.CHIP_NAME)
 
         if self.sync_stub_detected:
             print("Stub is already running. No upload is necessary.")
             return self.STUB_CLASS(self)
 
+        secure_boot_workflow = (
+            self.CHIP_NAME == "ESP32-S3" and self.get_secure_boot_enabled()
+        )
+
         # Upload
         print("Uploading stub...")
         for field in [stub.text, stub.data]:
@@ -964,8 +1229,24 @@ class ESPLoader(object):
                     from_offs = seq * self.ESP_RAM_BLOCK
                     to_offs = from_offs + self.ESP_RAM_BLOCK
                     self.mem_block(field[from_offs:to_offs], seq)
+
         print("Running stub...")
-        self.mem_finish(stub.entry)
+        if not secure_boot_workflow:
+            self.mem_finish(stub.entry)
+        else:
+            # Bug in ESP32-S3 ROM prevents stub execution if secure boot is enabled
+            # Hijack the `read` function in ROM to point to the stub entrypoint
+            # got with GDB - p &rom_spiflash_legacy_funcs.read
+            rom_spiflash_legacy_funcs_read_ptr = 0x3FCEF688
+            self.mem_finish(0)  # Finish uploading to RAM but don't run the stub yet
+            stored_read_pointer = self.read_reg(rom_spiflash_legacy_funcs_read_ptr)
+            self.write_reg(rom_spiflash_legacy_funcs_read_ptr, stub.entry)
+            self.command(  # Trigger the `read` in ROM to jump to the stub entrypoint
+                self.ESP_READ_FLASH_SLOW,
+                struct.pack("<II", 0, 0),
+                wait_response=False,
+            )
+
         try:
             p = self.read()
         except StopIteration:
@@ -977,6 +1258,9 @@ class ESPLoader(object):
 
         if p != b"OHAI":
             raise FatalError(f"Failed to start stub. Unexpected response: {p}")
+        if secure_boot_workflow:
+            # Restore the original `read` function pointer
+            self.write_reg(rom_spiflash_legacy_funcs_read_ptr, stored_read_pointer)
         print("Stub running...")
         return self.STUB_CLASS(self)
 
@@ -1059,20 +1343,21 @@ class ESPLoader(object):
     def flash_md5sum(self, addr, size):
         # the MD5 command returns additional bytes in the standard
         # command reply slot
+        RESP_DATA_LEN = 32
+        RESP_DATA_LEN_STUB = 16
         timeout = timeout_per_mb(MD5_TIMEOUT_PER_MB, size)
         res = self.check_command(
             "calculate md5sum",
             self.ESP_SPI_FLASH_MD5,
             struct.pack("<IIII", addr, size, 0, 0),
+            resp_data_len=RESP_DATA_LEN_STUB if self.IS_STUB else RESP_DATA_LEN,
             timeout=timeout,
         )
 
-        if len(res) == 32:
+        if not self.IS_STUB:
             return res.decode("utf-8")  # already hex formatted
-        elif len(res) == 16:
-            return hexify(res).lower()
         else:
-            raise FatalError("MD5Sum command returned unexpected result: %r" % res)
+            return hexify(res).lower()
 
     @stub_and_esp32_function_only
     def change_baud(self, baud):
@@ -1094,10 +1379,6 @@ class ESPLoader(object):
 
     @stub_function_only
     def erase_region(self, offset, size):
-        if offset % self.FLASH_SECTOR_SIZE != 0:
-            raise FatalError("Offset to erase from must be a multiple of 4096")
-        if size % self.FLASH_SECTOR_SIZE != 0:
-            raise FatalError("Size of data to erase must be a multiple of 4096")
         timeout = timeout_per_mb(ERASE_REGION_TIMEOUT_PER_MB, size)
         self.check_command(
             "erase region",
@@ -1302,7 +1583,9 @@ class ESPLoader(object):
         self.write_reg(
             SPI_USR2_REG, (7 << SPI_USR2_COMMAND_LEN_SHIFT) | spiflash_command
         )
-        if addr and addr_len > 0:
+        if addr_len > 0:
+            if self.SPI_ADDR_REG_MSB:
+                addr = addr << (32 - addr_len)
             self.write_reg(SPI_ADDR_REG, addr)
         if data_bits == 0:
             self.write_reg(SPI_W0_REG, 0)  # clear data register before we read it
@@ -1407,7 +1690,12 @@ class ESPLoader(object):
         #   See the self.XTAL_CLK_DIVIDER parameter for this factor.
         uart_div = self.read_reg(self.UART_CLKDIV_REG) & self.UART_CLKDIV_MASK
         est_xtal = (self._port.baudrate * uart_div) / 1e6 / self.XTAL_CLK_DIVIDER
-        norm_xtal = 40 if est_xtal > 33 else 26
+        if est_xtal > 45:
+            norm_xtal = 48
+        elif est_xtal > 33:
+            norm_xtal = 40
+        else:
+            norm_xtal = 26
         if abs(norm_xtal - est_xtal) > 1:
             print(
                 "WARNING: Detected crystal freq %.2fMHz is quite different to "
@@ -1416,9 +1704,13 @@ class ESPLoader(object):
             )
         return norm_xtal
 
-    def hard_reset(self):
+    def hard_reset(self, uses_usb=False):
         print("Hard resetting via RTS pin...")
-        HardReset(self._port)()
+        cfg_custom_hard_reset_sequence = cfg.get("custom_hard_reset_sequence")
+        if cfg_custom_hard_reset_sequence is not None:
+            CustomReset(self._port, cfg_custom_hard_reset_sequence)()
+        else:
+            HardReset(self._port, uses_usb)()
 
     def soft_reset(self, stay_in_bootloader):
         if not self.IS_STUB:
@@ -1443,22 +1735,12 @@ class ESPLoader(object):
                 # in the stub loader
                 self.command(self.ESP_RUN_USER_CODE, wait_response=False)
 
-    def check_chip_id(self):
-        try:
-            chip_id = self.get_chip_id()
-            if chip_id != self.IMAGE_CHIP_ID:
-                print(
-                    "WARNING: Chip ID {} ({}) doesn't match expected Chip ID {}. "
-                    "esptool may not work correctly.".format(
-                        chip_id,
-                        self.UNSUPPORTED_CHIPS.get(chip_id, "Unknown"),
-                        self.IMAGE_CHIP_ID,
-                    )
-                )
-                # Try to flash anyways by disabling stub
-                self.stub_is_disabled = True
-        except NotImplementedInROMError:
-            pass
+    def watchdog_reset(self):
+        print(
+            f"WARNING: Watchdog hard reset is not supported on {self.CHIP_NAME}, "
+            "attempting classic hard reset instead."
+        )
+        self.hard_reset()
 
 
 def slip_reader(port, trace_function):
@@ -1589,12 +1871,14 @@ class HexFormatter(object):
             while len(s) > 0:
                 line = s[:16]
                 ascii_line = "".join(
-                    c
-                    if (
-                        c == " "
-                        or (c in string.printable and c not in string.whitespace)
+                    (
+                        c
+                        if (
+                            c == " "
+                            or (c in string.printable and c not in string.whitespace)
+                        )
+                        else "."
                     )
-                    else "."
                     for c in line.decode("ascii", "replace")
                 )
                 s = s[16:]
