@@ -63,6 +63,12 @@ if arg_chip not in SUPPORTED_CHIPS:
 print(f"\nHost tests of espefuse.py for {arg_chip}:")
 print("Running espefuse.py tests...")
 
+# The default value of the program name for argparse has changed in Python 3.14
+# https://docs.python.org/dev/whatsnew/3.14.html#argparse
+ESPEFUSE_MODNAME = (
+    "__main__.py" if sys.version_info < (3, 14) else "python3 -m espefuse"
+)
+
 
 @pytest.mark.host_test
 class EfuseTestCase:
@@ -173,11 +179,13 @@ class EfuseTestCase:
 
 class TestReadCommands(EfuseTestCase):
     def test_help(self):
-        self.espefuse_not_virt_py("--help", check_msg="usage: __main__.py [-h]")
+        self.espefuse_not_virt_py("--help", check_msg=f"usage: {ESPEFUSE_MODNAME} [-h]")
         self.espefuse_not_virt_py(f"--chip {arg_chip} --help")
 
     def test_help2(self):
-        self.espefuse_not_virt_py("", check_msg="usage: __main__.py [-h]", ret_code=1)
+        self.espefuse_not_virt_py(
+            "", check_msg=f"usage: {ESPEFUSE_MODNAME} [-h]", ret_code=1
+        )
 
     def test_dump(self):
         self.espefuse_py("dump -h")
@@ -211,9 +219,6 @@ class TestReadCommands(EfuseTestCase):
             ret_code=2,
         )
 
-    @pytest.mark.skipif(
-        arg_chip == "esp32p4", reason="No Custom MAC Address defined yet"
-    )
     def test_get_custom_mac(self):
         self.espefuse_py("get_custom_mac -h")
         if arg_chip == "esp32":
@@ -233,7 +238,7 @@ class TestReadCommands(EfuseTestCase):
             self.espefuse_py("burn_efuse BLK_VERSION_MAJOR 1")
         elif arg_chip in ["esp32c2", "esp32s2", "esp32c6"]:
             self.espefuse_py("burn_efuse BLK_VERSION_MINOR 1")
-        elif arg_chip in ["esp32h2", "esp32h2beta1"]:
+        elif arg_chip in ["esp32h2", "esp32h2beta1", "esp32p4"]:
             self.espefuse_py("burn_efuse BLK_VERSION_MINOR 2")
         self.espefuse_py("adc_info")
 
@@ -280,6 +285,9 @@ class TestReadProtectionCommands(EfuseTestCase):
         output = self.espefuse_py(cmd)
         assert count_protects == output.count("is already read protected")
 
+    @pytest.mark.skipif(
+        arg_chip == "esp32p4", reason="BLOCK_SYS_DATA2 is used by ADC calib"
+    )
     def test_read_protect_efuse2(self):
         self.espefuse_py("write_protect_efuse RD_DIS")
         if arg_chip == "esp32":
@@ -324,7 +332,8 @@ class TestReadProtectionCommands(EfuseTestCase):
         else:
             key1_purpose = (
                 "USER"
-                if arg_chip in ["esp32p4", "esp32c61", "esp32c5", "esp32c5beta3"]
+                if arg_chip
+                in ["esp32p4", "esp32c61", "esp32c5", "esp32c5beta3", "esp32h4"]
                 else "RESERVED"
             )
             self.espefuse_py(
@@ -398,7 +407,7 @@ class TestWriteProtectionCommands(EfuseTestCase):
                            SPI_BOOT_CRYPT_CNT"""
             efuse_lists2 = "RD_DIS KEY_PURPOSE_0 KEY_PURPOSE_2"
         else:
-            efuse_lists = """RD_DIS DIS_ICACHE DIS_FORCE_DOWNLOAD
+            efuse_lists = """RD_DIS DIS_FORCE_DOWNLOAD
                            DIS_DOWNLOAD_MANUAL_ENCRYPT
                            USB_EXCHG_PINS WDT_DELAY_SEL SPI_BOOT_CRYPT_CNT
                            SECURE_BOOT_KEY_REVOKE0 SECURE_BOOT_KEY_REVOKE1
@@ -417,16 +426,18 @@ class TestWriteProtectionCommands(EfuseTestCase):
                 "esp32c61",
                 "esp32c5",
                 "esp32c5beta3",
+                "esp32h21",
+                "esp32h4",
             ]:
                 efuse_lists += """ DIS_DOWNLOAD_ICACHE
                             SPI_PAD_CONFIG_CLK SPI_PAD_CONFIG_Q
                             SPI_PAD_CONFIG_D SPI_PAD_CONFIG_CS SPI_PAD_CONFIG_HD
                             SPI_PAD_CONFIG_WP SPI_PAD_CONFIG_DQS SPI_PAD_CONFIG_D4
                             SPI_PAD_CONFIG_D5 SPI_PAD_CONFIG_D6 SPI_PAD_CONFIG_D7"""
-            efuse_lists2 = "RD_DIS DIS_ICACHE"
+            efuse_lists2 = "RD_DIS"
         self.espefuse_py(f"write_protect_efuse {efuse_lists}")
         output = self.espefuse_py(f"write_protect_efuse {efuse_lists2}")
-        assert output.count("is already write protected") == 2
+        assert output.count("is already write protected") >= 1
 
     def test_write_protect_efuse2(self):
         if arg_chip == "esp32":
@@ -439,7 +450,6 @@ class TestWriteProtectionCommands(EfuseTestCase):
             )
 
 
-@pytest.mark.skipif(arg_chip == "esp32p4", reason="No Custom MAC Address defined yet")
 class TestBurnCustomMacCommands(EfuseTestCase):
     def test_burn_custom_mac(self):
         self.espefuse_py("burn_custom_mac -h")
@@ -682,9 +692,6 @@ class TestBurnEfuseCommands(EfuseTestCase):
         assert "(Override SD_CMD pad (GPIO11/SPICS0)) 0b00000 -> 0b11111" in output
         assert "BURN BLOCK0  - OK (all write block bits are set)" in output
 
-    @pytest.mark.skipif(
-        arg_chip == "esp32p4", reason="No Custom MAC Address defined yet"
-    )
     def test_burn_mac_custom_efuse(self):
         crc_msg = "(OK)"
         self.espefuse_py("burn_efuse -h")
@@ -733,7 +740,13 @@ class TestBurnEfuseCommands(EfuseTestCase):
                 SECURE_BOOT_EN 1 \
                 UART_PRINT_CONTROL 1"
             )
-            if arg_chip not in ["esp32c5", "esp32c5beta3", "esp32c61"]:
+            if arg_chip not in [
+                "esp32c5",
+                "esp32c5beta3",
+                "esp32c61",
+                "esp32h21",
+                "esp32h4",
+            ]:
                 # chips having the OPTIONAL_UNIQUE_ID field
                 self.espefuse_py(
                     "burn_efuse \
@@ -1286,13 +1299,15 @@ class TestBurnBlockDataCommands(EfuseTestCase):
         ) in output
         self.check_data_block_in_log(output, f"{IMAGES_DIR}/256bit")
 
-        self.espefuse_py(
-            f"burn_block_data \
-            BLOCK10 {IMAGES_DIR}/256bit_1"
-        )
-        self.check_data_block_in_log(
-            self.espefuse_py("summary -d"), f"{IMAGES_DIR}/256bit_1"
-        )
+        if arg_chip != "esp32p4":
+            # BLOCK10 is free. In P4 it is used for ADC calib data.
+            self.espefuse_py(
+                f"burn_block_data \
+                BLOCK10 {IMAGES_DIR}/256bit_3"
+            )
+            self.check_data_block_in_log(
+                self.espefuse_py("summary -d"), f"{IMAGES_DIR}/256bit_3"
+            )
 
         self.espefuse_py(
             f"burn_block_data \
@@ -1306,7 +1321,7 @@ class TestBurnBlockDataCommands(EfuseTestCase):
             in output
         )
         self.check_data_block_in_log(output, f"{IMAGES_DIR}/256bit")
-        self.check_data_block_in_log(output, f"{IMAGES_DIR}/256bit_1", 2)
+        self.check_data_block_in_log(output, f"{IMAGES_DIR}/256bit_1")
         self.check_data_block_in_log(output, f"{IMAGES_DIR}/256bit_2")
 
     def test_burn_block_data_check_errors(self):
@@ -1862,7 +1877,7 @@ class TestExecuteScriptsCommands(EfuseTestCase):
         os.chdir(self.stored_dir)
 
     @pytest.mark.skipif(
-        arg_chip in ["esp32c2", "esp32p4"],
+        arg_chip in ["esp32c2", "esp32p4", "esp32h21", "esp32h4"],
         reason="These chips do not have eFuses used in this test",
     )
     def test_execute_scripts_with_check_that_only_one_burn(self):
@@ -1872,7 +1887,7 @@ class TestExecuteScriptsCommands(EfuseTestCase):
         self.espefuse_py("execute_scripts execute_efuse_script2.py")
 
     @pytest.mark.skipif(
-        arg_chip in ["esp32c2", "esp32p4"],
+        arg_chip in ["esp32c2", "esp32p4", "esp32h21", "esp32h4"],
         reason="These chips do not have eFuses used in this test",
     )
     def test_execute_scripts_with_check(self):
@@ -1954,17 +1969,17 @@ class TestMultipleCommands(EfuseTestCase):
 
         self.espefuse_py(
             f"-h {command1} {command2}",
-            check_msg="usage: __main__.py [-h]",
+            check_msg=f"usage: {ESPEFUSE_MODNAME} [-h]",
         )
 
         self.espefuse_py(
             f"{command1} -h {command2}",
-            check_msg="usage: __main__.py burn_key_digest [-h]",
+            check_msg=f"usage: {ESPEFUSE_MODNAME} burn_key_digest [-h]",
         )
 
         self.espefuse_py(
             f"{command1} {command2} -h",
-            check_msg="usage: __main__.py burn_key [-h]",
+            check_msg=f"usage: {ESPEFUSE_MODNAME} burn_key [-h]",
         )
 
     @pytest.mark.skipif(
